"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tokenizer_1 = require("../../sucrase-babylon/tokenizer");
const Transformer_1 = require("./Transformer");
/**
 * Implementation of babel-plugin-transform-react-display-name, which adds a
 * display name to usages of React.createClass and createReactClass.
 *
 * This implementation has the following limitations compared with the
 * - It does not handle `export default React.createClass`, using the filename,
 *   since Sucrase currently does not know the name of the current file.
 */
class ReactDisplayNameTransformer extends Transformer_1.default {
    constructor(rootTransformer, tokens, importProcessor) {
        super();
        this.rootTransformer = rootTransformer;
        this.tokens = tokens;
        this.importProcessor = importProcessor;
    }
    process() {
        const startIndex = this.tokens.currentIndex();
        if (this.tokens.matchesContextual(32 /* _createReactClass */)) {
            const newName = this.importProcessor.getIdentifierReplacement("createReactClass");
            if (newName) {
                this.tokens.replaceToken(`(0, ${newName})`);
            }
            else {
                this.tokens.copyToken();
            }
            this.tryProcessCreateClassCall(startIndex);
            return true;
        }
        if (this.tokens.matches3(2048 /* name */, 9216 /* dot */, 2048 /* name */) &&
            this.tokens.matchesContextual(30 /* _React */) &&
            this.tokens.matchesContextualAtIndex(this.tokens.currentIndex() + 2, 31 /* _createClass */)) {
            const newName = this.importProcessor.getIdentifierReplacement("React");
            if (newName) {
                this.tokens.replaceToken(newName);
                this.tokens.copyToken();
                this.tokens.copyToken();
            }
            else {
                this.tokens.copyToken();
                this.tokens.copyToken();
                this.tokens.copyToken();
            }
            this.tryProcessCreateClassCall(startIndex);
            return true;
        }
        return false;
    }
    /**
     * This is called with the token position at the open-paren.
     */
    tryProcessCreateClassCall(startIndex) {
        const displayName = this.findDisplayName(startIndex);
        if (!displayName) {
            return;
        }
        if (this.classNeedsDisplayName()) {
            this.tokens.copyExpectedToken(6144 /* parenL */);
            this.tokens.copyExpectedToken(4096 /* braceL */);
            this.tokens.appendCode(`displayName: '${displayName}',`);
            this.rootTransformer.processBalancedCode();
            this.tokens.copyExpectedToken(5120 /* braceR */);
            this.tokens.copyExpectedToken(6656 /* parenR */);
        }
    }
    findDisplayName(startIndex) {
        if (this.tokens.matchesAtIndex(startIndex - 2, [2048 /* name */, 14368 /* eq */]) &&
            !this.tokens.matchesAtIndex(startIndex - 3, [9216 /* dot */])) {
            // This is an assignment (or declaration) with an identifier LHS, so use
            // that identifier name.
            return this.tokens.identifierNameAtIndex(startIndex - 2);
        }
        if (this.tokens.tokens[startIndex - 2].identifierRole === tokenizer_1.IdentifierRole.ObjectKey) {
            // This is an object literal value.
            return this.tokens.identifierNameAtIndex(startIndex - 2);
        }
        return null;
    }
    /**
     * We only want to add a display name when this is a function call containing
     * one argument, which is an object literal without `displayName` as an
     * existing key.
     */
    classNeedsDisplayName() {
        let index = this.tokens.currentIndex();
        if (!this.tokens.matches2(6144 /* parenL */, 4096 /* braceL */)) {
            return false;
        }
        // The block starts on the {, and we expect any displayName key to be in
        // that context. We need to ignore other other contexts to avoid matching
        // nested displayName keys.
        const objectStartIndex = index + 1;
        const objectContextId = this.tokens.tokens[objectStartIndex].contextId;
        if (objectContextId == null) {
            throw new Error("Expected non-null context ID on object open-brace.");
        }
        for (; index < this.tokens.tokens.length; index++) {
            const token = this.tokens.tokens[index];
            if (token.type === 5120 /* braceR */ && token.contextId === objectContextId) {
                index++;
                break;
            }
            if (this.tokens.matchesContextualAtIndex(index, 33 /* _displayName */) &&
                this.tokens.tokens[index].identifierRole === tokenizer_1.IdentifierRole.ObjectKey &&
                token.contextId === objectContextId) {
                // We found a displayName key, so bail out.
                return false;
            }
        }
        if (index === this.tokens.tokens.length) {
            throw new Error("Unexpected end of input when processing React class.");
        }
        // If we got this far, we know we have createClass with an object with no
        // display name, so we want to proceed as long as that was the only argument.
        return (this.tokens.matchesAtIndex(index, [6656 /* parenR */]) ||
            this.tokens.matchesAtIndex(index, [7168 /* comma */, 6656 /* parenR */]));
    }
}
exports.default = ReactDisplayNameTransformer;
